/**
 * Copyright (c) 2011 Generative Software Development Lab, 
 * University of Waterloo, Canada. 
 * All Rights Reserved.
 * ----------------------------------------------------------
 *           \                                                                           
 *            \                                                                          
 *               __                                                                      
 *          w  c(..)o                                                                    
 *           \__(o)                                                                      
 *               /\                                                                      
 *            w_/(_)-~                                                                   
 *                /|                                                                     
 *               | \                                                                     
 *               m  m  
 */
package org.apache.bcel.generic;

import java.util.List;

import org.apache.bcel.classfile.AccessFlags;
import org.apache.bcel.classfile.Attribute;
import org.apache.bcel.classfile.Field;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.support.PushedDown;
import org.apache.bcel.support.WrapperMember;
import org.apache.bcel.support.WrapperRecycler;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.tree.ClassNode;
import org.objectweb.asm.tree.FieldNode;
import org.objectweb.asm.tree.InnerClassNode;
import org.objectweb.asm.tree.MethodNode;

/**
 * 
 * TODO Comment!!
 * 
 * @author Thiago Tonelli Bartolomei <ttonelli@uwaterloo.ca>
 */
public class ClassGen extends AccessFlags {

	/**
	 * Wrappee field.
	 */
	@WrapperMember
	private ClassNode classNode = null; 
	
	/**
	 * Stateful adapter.
	 */
	@WrapperMember
	private ConstantPoolGen cp = null;
	
	/**
	 * Wrapper constructor.
	 *
	 * @param node
	 */
	@WrapperMember
	ClassGen(ClassNode node) {
		setClassNode(node);
	}

	/**
	 * Sets the class node wrappee, registering in the identity map.
	 *
	 * @param node
	 */
	@WrapperMember
	protected ClassNode setClassNode(ClassNode node) {
		this.classNode = node;
		WrapperRecycler.add(this, node);
		return node;
	}
	
	/**
	 * Getter allows for wrappee refinement.
	 *
	 * @return
	 */
	@WrapperMember
	protected ClassNode getClassNode() {
		return classNode;
	}
	
	/**
	 * Creates a class gen based on this java class.
	 *
	 * @param prototype
	 */
	@SuppressWarnings("unchecked")
	public ClassGen(JavaClass prototype) {
		
		ClassNode orig = prototype.getClassNode();
		
		// clone the class node
		ClassNode classNode = new ClassNode();
		classNode.name = orig.name;
		classNode.superName = orig.superName;
		classNode.access = orig.access;
		classNode.interfaces.addAll(orig.interfaces);
		classNode.version = orig.version;
        classNode.sourceFile = orig.sourceFile;
        classNode.sourceDebug = orig.sourceDebug;
        classNode.outerClass = orig.outerClass;
        classNode.outerMethod = orig.outerMethod;
        classNode.outerMethodDesc = orig.outerMethodDesc;
		
        this.setClassNode(classNode);
        
        // create a new constant pool
        this.cp = new ConstantPoolGen();
        
		// clone the methods
        for (Object o : orig.methods) {
        	MethodNode node = (MethodNode) o;
        	classNode.methods.add(node);
        }
        
        // copy the fields
        for (Object o : orig.fields) {
        	FieldNode node = (FieldNode) o;
        	this.addField(new Field(new FieldNode(node.access, node.name, node.desc, node.signature, node.value)));
        }
        
        // copy inner classes
        for (Object o : orig.innerClasses) {
        	InnerClassNode inner = (InnerClassNode) o;
        	classNode.innerClasses.add(new InnerClassNode(inner.name, inner.outerName, inner.innerName, inner.access));
        }
	}

	/**
	 * Creates a new ClassGen object with these values.
	 *
	 * @param class_name
	 * @param super_class_name
	 * @param file_name
	 * @param access_flags
	 * @param interfaces
	 */
	@SuppressWarnings("unchecked")
	public ClassGen(
			String class_name, 
			String super_class_name, 
			String file_name, 
			int access_flags,
            String[] interfaces) {
		
        this.cp = new ConstantPoolGen();
        
		ClassNode classNode = this.setClassNode(new ClassNode());
		classNode.name = Type.getInternalName(class_name);
		classNode.superName = Type.getInternalName(super_class_name);
		classNode.access = access_flags;
		if (null != interfaces)
			for (int i = 0; i < interfaces.length; i++)
				classNode.interfaces.add(Type.getInternalName(interfaces[i]));
		
		// by default, BCEL uses 45/3 for major/minor
		classNode.version = (3 << 16) | 45;
		
		// set the source filename
	    if (null != file_name)
        	classNode.sourceFile = file_name;
	}
	
	/**
	 * Adds the method to this class.
	 *
	 * @param arg0
	 */
	@SuppressWarnings("unchecked")
	public void addMethod(Method method) {
		getClassNode().methods.add(method.getNode());
	}

	/**
	 * Adds a new attribute to the class.
	 *
	 * @param a
	 */
	public void addAttribute(Attribute a) {
		throw new UnsupportedOperationException();
    }

	/**
	 * Replaces a method in this class
	 *
	 * @param oldMethod
	 * @param newMethod
	 */
	@SuppressWarnings("unchecked")
	public void replaceMethod(Method oldMethod, Method newMethod) {
		ClassNode n = getClassNode();
		int i = n.methods.indexOf(oldMethod.getNode());
		if (i >= 0) {
			n.methods.remove(oldMethod.getNode());
			n.methods.add(i, newMethod.getNode());
		}
	}

	/**
	 * Gets this class' constant pool.
	 *
	 * @return
	 */
	public ConstantPoolGen getConstantPool() {
		return cp;
	}

	/**
	 * Gets the name of this class.
	 *
	 * @return
	 */
	public String getClassName() {
		return Type.getClassName(getClassNode().name);
	}

	/**
	 * Gets the method at this index.
	 *
	 * @param index
	 * @return
	 */
	public Method getMethodAt(int index) {
		return WrapperRecycler.wrap(getClassNode().methods.get(index), Method.class);
	}

	/**
	 * Replaces the method at this index with this new method.
	 *
	 * @param method
	 * @param index
	 */
	@SuppressWarnings("unchecked")
	public void setMethodAt(Method method, int index) {
		getClassNode().methods.set(index, method.getNode());
	}

	public Field containsField(String name) {
		for (Object n : this.getClassNode().fields) 
			if (name.equals(((FieldNode) n).name))
				return WrapperRecycler.wrap(n, Field.class);
				
		return null;
	}

	/**
	 * Return a new JavaClass that encapsulates the same information
	 * as this class.
	 *
	 * @return
	 */
	public JavaClass getJavaClass() {
		return new JavaClass(this.getClassNode());
	}

	/**
	 * Adds a field to this class.
	 *
	 * @param field
	 */
	@SuppressWarnings("unchecked")
	public void addField(Field field) {
		getClassNode().fields.add(field.getNode());
	}

	/**
	 * Gets the methods of this class.
	 *
	 * @return
	 */
	public Method[] getMethods() {
		List<?> methods = getClassNode().methods;
		Method[] array = new Method[methods.size()];
		for (int i = 0; i < array.length; i++) {
			array[i] = WrapperRecycler.wrap(methods.get(i), Method.class);
		}
		return array;
	}
	
	/**
	 * Originally on super type.
	 *
	 * @return
	 */
	@Override
	@PushedDown
	public boolean isPublic() {
		return (getClassNode().access & Opcodes.ACC_PUBLIC) != 0;
	}

	/**
	 * Originally on super type.
	 *
	 * @param flags
	 */
	@Override
	@PushedDown
	public void setAccessFlags(int flags) {
		getClassNode().access = flags;
	}

	/**
	 * Originally on super type.
	 *
	 * @return
	 */
	@Override
	@PushedDown
	public int getAccessFlags() {
		return getClassNode().access;
	}
	
	/**
	 * Computes the major portion of the class version.
	 *
	 * @return
	 */
	public int getMajor() {
		throw new UnsupportedOperationException();
    }
    
	/**
	 * Computes the minor portion of the class version.
	 *
	 * @return
	 */
	public int getMinor() {
		throw new UnsupportedOperationException();
    }
	
	/**
	 * Gets the class super class name.
	 *
	 * @return
	 */
	public String getSuperclassName() {
		return Type.getClassName(getClassNode().superName);
    }

	/**
	 * Gets the name of the source file of this class.
	 *
	 * @return
	 */
    public String getFileName() {
    	return Type.getClassName(getClassNode().sourceFile);
    }
    
    public int getSuperclassNameIndex() {
    	throw new UnsupportedOperationException();
    }


    public int getClassNameIndex() {
    	throw new UnsupportedOperationException();
    }
    
    /**
     * Gets the fields of this class.
     *
     * @return
     */
    public Field[] getFields() {
    	throw new UnsupportedOperationException();
    }

    /**
     * Gets the attributes of this class.
     *
     * @return
     */
    public Attribute[] getAttributes() {
    	throw new UnsupportedOperationException();
    }
    
    /**
     * Gets the list of internal names of the interfaces directly implemented
     * by this class.
     *
     * @return
     */
	public String[] getInterfaceNames() {
		List<?> internalNames = getClassNode().interfaces;
		String[] interfaces = new String[internalNames.size()];
		for (int i = 0; i < interfaces.length; i++)
			interfaces[i] = Type.getClassName((String) internalNames.get(i)); 
		return interfaces;
    }
}
