/**
 * Copyright (c) 2011 Generative Software Development Lab, 
 * University of Waterloo, Canada. 
 * All Rights Reserved.
 * ----------------------------------------------------------
 *           \                                                                           
 *            \                                                                          
 *               __                                                                      
 *          w  c(..)o                                                                    
 *           \__(o)                                                                      
 *               /\                                                                      
 *            w_/(_)-~                                                                   
 *                /|                                                                     
 *               | \                                                                     
 *               m  m  
 */
package org.apache.bcel.generic;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.apache.bcel.support.WrapperMember;
import org.apache.bcel.support.WrapperRecycler;
import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.LabelNode;


/**
 * This is a top of an inheritance hierarchy of instructions.
 * 
 * In ASM there's a single type, AbstractInsnNode, so all sub-types of this
 * class share the same wrappee, but configure it differently.
 * 
 * @author Thiago Tonelli Bartolomei <ttonelli@uwaterloo.ca>
 */
public abstract class Instruction {
	
	/**
	 * Wrappee field.
	 */
	@WrapperMember
	private AbstractInsnNode node = null;
	
	Instruction() {
	}
	
	/**
	 * Wrapper constructor.
	 *
	 * @param node
	 */
	@WrapperMember
	Instruction(AbstractInsnNode node) {
		setNode(node);
	}

	/**
	 * Sets the class node wrappee, registering in the identity map.
	 *
	 * @param node
	 */
	@WrapperMember
	protected AbstractInsnNode setNode(AbstractInsnNode node) {
		this.node = node;
		WrapperRecycler.add(this, node);
		return node;
	}
	
	/**
	 * Getter allows for wrappee refinement.
	 *
	 * @return
	 */
	@WrapperMember
	protected AbstractInsnNode getNode() {
		return node;
	}

	/**
	 * The list of handles that target this instruction.
	 */
	@WrapperMember
	List<InstructionHandle> handles = null;
	
	/**
	 * Creates a new handle targeting this instruction.
	 * 
	 * This method will guarantee that the handle points to a correct ASM
	 * node, which can be either the one pointed initially by this instruction
	 * or a clone of that node.
	 *
	 * @return
	 */
	@WrapperMember
	public InstructionHandle createHandle() {
		AbstractInsnNode node = (null == handles) ? getNode() : this.cloneNode();
		return addHandle(createHandle(node));
	}
	
	@WrapperMember
	public InstructionHandle addHandle(InstructionHandle handle) {
		if (null == handles)
			handles = new ArrayList<InstructionHandle>(8);
		handles.add(handle);
		handle.setInstruction(this);
		return handle;
	}
	
	@WrapperMember
	public void removeHandle(InstructionHandle handle) {
		if (null != handles) {
			handles.remove(handle);
			if (handles.isEmpty())
				handles = null;
		}
	}
	
	@WrapperMember
	protected InstructionHandle getHandleFromList(InstructionList list) {
		for (InstructionHandle ih : handles)
			if (ih.list == list)
				return ih;
		return null;
	}
	
	@WrapperMember
	InstructionHandle createHandle(AbstractInsnNode node) {
		return InstructionHandle.createHandle(node);
	}

	@WrapperMember
	public static Instruction createInstruction(AbstractInsnNode node) {
		int opcode = node.getOpcode();
		// first see if we can use a singleton
		if (InstructionConstants.INSTRUCTIONS[opcode] != null)
			return InstructionConstants.INSTRUCTIONS[opcode];
		
		// now instantiate a new instruction
		if (InstructionConstants.INSTRUCTION_WRAPPERS[opcode] != null)
			try {
				Instruction i = (Instruction) InstructionConstants.INSTRUCTION_WRAPPERS[opcode].newInstance();
				i.setNode(node);
				return i;
			} catch (Exception e) {
				e.printStackTrace();
				throw new RuntimeException("Could not instantiate " + opcode + " : " + e);
			}
		
		throw new RuntimeException("Instruction with opcode " + opcode + " is not implemented yet!");
	}

	/**
	 * Rewrites the labels of this node using this map from old labels to new instruction
	 * handles.
	 * 
	 * This is called when a list is being cloned. Sub-types must rewrite all uses of labels
	 * in the node to the labels of the corresponding instruction handle in the map.
	 *
	 * @param node
	 * @param map
	 * @return
	 */
	@WrapperMember
	Instruction rewrite(AbstractInsnNode node, Map<LabelNode, InstructionHandle> map) {
		return this;
	}
	
	/**
	 * Gets the instruction's opcode.
	 *
	 * @return
	 */
	public int getOpcode() {
		return getNode().getOpcode();
	}
	
	/**
	 * Clones this instruction, if needed.
	 * 
	 * Some instructions have singletons in InstructionConstants, and they will be used.
	 *
	 * @return
	 */
	@WrapperMember
	public Instruction clone() {
		if (InstructionConstants.INSTRUCTIONS[this.getOpcode()] != null) {
			return InstructionConstants.INSTRUCTIONS[this.getOpcode()];
		}
		return cloneInstruction();
	}
	
	/**
	 * Shallow clone of this instruction.
	 * 
	 * Must return a new object of the same type as this instruction, 
	 * and it must be initialized with a cloned node.
	 *
	 * @return
	 */
	@WrapperMember
	abstract Instruction cloneInstruction();
	
	/**
	 * Returns a clone of the instruction's target node.
	 * 
	 * The clone can use the same original labels as the original node.
	 *
	 * @return
	 */
	@WrapperMember
	abstract AbstractInsnNode cloneNode();
	
	
	/**
	 * 
	 * TODO Comment!!
	 *
	 * @param verbose
	 * @return
	 */
	public String toString(boolean verbose) {
		throw new UnsupportedOperationException();
	}
}
