/**
 * Copyright (c) 2009-2011 Generative Software Development Lab, 
 * University of Waterloo, Canada. 
 * All Rights Reserved.
 * ----------------------------------------------------------
 *           \                                                                           
 *            \                                                                          
 *               __                                                                      
 *          w  c(..)o                                                                    
 *           \__(o)                                                                      
 *               /\                                                                      
 *            w_/(_)-~                                                                   
 *                /|                                                                     
 *               | \                                                                     
 *               m  m  
 */
package org.apache.bcel.support;

import java.lang.ref.WeakReference;
import java.lang.reflect.Constructor;
import java.util.IdentityHashMap;
import java.util.Map;

/**
 * A generic wrapper identity map implementation, originally from xom2jdom.
 * 
 * @author Thiago T. Bartolomei
 */
public class WrapperRecycler {

	/**
	 * API packages.
	 */
	public static String[] apis = new String[] {};
	
	/**
	 * Maps a wrappee to a reference to its wrappers.
	 * 
	 * @see WrapperReference
	 */
	protected static Map<Object, WrapperReference> map = 
			new IdentityHashMap<Object, WrapperReference>();
	
	/**
	 * Registers the relationship between this wrappee and this wrapper.
	 *
	 * @param wrapper
	 * @param wrappee
	 */
	public static void add(Object wrapper, Object wrappee) {
		if (! contains(wrapper, wrappee)) {
			WrapperReference ref = null;
			if (isExtension(wrapper.getClass().getName())) {
				ref = new WrapperReference(wrapper);
			} else {
				ref = new WrapperReference(new WeakReference<Object>(wrapper));
			}
			// Always register the new ref, but point to the previous one if it existed before
			WrapperReference init = map.get(wrappee);
			map.put(wrappee, ref);
			if (null != init)
				ref.next = init;
		}
	}
	
	public static boolean contains(Object wrapper, Object wrappee) {
		if (null != wrappee) {
			WrapperReference ref = map.get(wrappee);
			
			// search in the linked list
			while (null != ref) {
				if (ref.get() == wrapper) {
					return true;
				}
				ref = ref.next;
			}
		}
		return false;
	}
	/**
	 * Explicitly removes the relationship between this wrapper and wrappee from the map.
	 * 
	 * This is useful if the wrapper object needs to be reused with a different wrappee.
	 *
	 * @param wrapper
	 * @param wrappee
	 * @return
	 */
	public static boolean remove(Object wrapper, Object wrappee) {
		if (null != wrappee) {
			WrapperReference init = map.get(wrappee);
			// couldn't find any reference with wrappee
			if (null == init)
				return false;
			
			// the first reference contains the wrapper
			if (init.get() == wrapper) {
				if (init.next == null)
					map.remove(wrappee);
				else
					map.put(wrappee, init.next);
				return true;
			}
			
			// the wrapper is not in the first reference, look for it
			WrapperReference ref = init;
			
			// clean all but init
			while(ref.next != null) {
				if (ref.next.get() == wrapper) {
					ref.next = ref.next.next;
					return true;
				}
				ref = ref.next;
			}
		}
		return false;
	}
	
	/**
	 * Tries to get the wrapper for this wrappee and return it.
	 * 
	 * If wrappee is null, this will promptly return null.
	 * 
	 * If there is no wrapper registered for the wrappee, this method
	 * will try to create a wrapper using the wrapper class. The class must 
	 * implement a single-parameter constructor that receives the wrappee.
	 *
	 * @param wrappee
	 * @param wrapper
	 * @return
	 */
	@SuppressWarnings("unchecked")
	public static <T> T wrap(Object wrappee, Class<T> wrapper) {
		if (null == wrappee) return null;
		
		// If the map does not know about this wrappee, create a wrapper
		if (! map.containsKey(wrappee))
			return create(wrappee, wrapper);
		
		// Clean the reference list from GC'ed references
		WrapperReference init = map.get(wrappee);
		WrapperReference ref = init;
		// clean all but init
		while(ref.next != null) {
			if (ref.next.get() == null)
				ref.next = ref.next.next;
			ref = ref.next;
		}
		// if init is cleaned, need to move it (or if there's no next, create)
		if (init.get() == null) {
			if (null == init.next)
				return create(wrappee, wrapper);
			init = init.next;
			map.put(wrappee, init);
		}
		
		// Now search for a suitable reference
		ref = init;
		while(ref != null) {
			Object o = ref.get();
			if (wrapper.isAssignableFrom(o.getClass()))
				return (T) o;
			ref = ref.next;
		}
		
		// Finally, if we could not find one with the appropriate type,
		// create and append to the list.
		return create(wrappee, wrapper);
	}
	
	/**
	 * Construct a new wrapper of this wrapper class, registers the relationship
	 * between the wrappee and wrapper, and returns the newly created wrapper.
	 *
	 * @param wrappee
	 * @param wrapper
	 * @return
	 */
	protected static <T> T create(Object wrappee, Class<T> wrapper) {
		T o = construct(wrapper, wrappee);
		add(o, wrappee);
		return o;
	}
	
	/**
	 * Constructs an object of the type c by calling its constructor that
	 * receives the wrappee as parameter. If any exceptions occur, return null.
	 *
	 * This method does not work in general, but if the wrappers follow the guidelines
	 * it will always work. That's because the constructor might return a supertype of
	 * T if T does not implement the wrapping constructor.
	 * 
	 * @param c
	 * @param wrappee
	 * @return
	 */
	@SuppressWarnings("unchecked")
	protected static <T> T construct(Class<T> c, Object wrappee) {
		try {
			Class<?> wrappeeClass = wrappee.getClass();
			Constructor<?> candidate = null;
			boolean needsDummyParameter = false;
			
			for(Constructor<?> constructor : c.getDeclaredConstructors()) {
				Class<?>[] params = constructor.getParameterTypes();
				// if we find a constructor that received the wrappee object, it
				// is a candidate
				if (params.length == 1) {
					if (params[0].isAssignableFrom(wrappeeClass)) { 
						candidate = constructor;
					}
					
				// if there's a constructor with the wrapper recycler as a dummy
				// parameter, it has precedence
				} else if (params.length == 2 && candidate == null) {
					if (params[1].isAssignableFrom(WrapperRecycler.class) &&
						params[0].isAssignableFrom(wrappeeClass)) {
						candidate = constructor;
						needsDummyParameter = true;
						break;
					}
				}
			}
			
			// If we found a candidate constructor, instantiate a wrapper
			if (null != candidate) {
				candidate.setAccessible(true);
				if (needsDummyParameter)
					return (T) candidate.newInstance(wrappee, null);
				else
					return (T) candidate.newInstance(wrappee);
			}
			
			System.err.println("Could not find appropriate constructor at " + c.getName());
			return null;
			
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		} 
	}
	
	/**
	 * Verifies if the qualifiedClassname belongs to a package that is
	 * part of the API definitions currently in use.
	 *
	 * If the class belongs to the API packages, then it is NOT an extension.
	 * Otherwise, it is.
	 * 
	 * @param qualifiedClassname
	 * @return
	 */
	protected static boolean isExtension(String qualifiedClassname) {
		for(String api : apis)
			if (qualifiedClassname.matches(api))
				return false;
		return true;
	}

	/**
	 * Keeps a reference to a wrapper.
	 * 
	 * The reference can be either weak or strong, but never both at the same time.
	 * 
	 * @author Thiago T. Bartolomei
	 */
	public static class WrapperReference {
		
		/**
		 * A weak reference to a wrapper.
		 */
		private final WeakReference<Object> weak;
		
		/**
		 * A strong reference to a wrapper.
		 */
		private final Object strong;
		
		/**
		 * A linked list pointer leading to the next wrapper reference.
		 */
		public WrapperReference next = null;
		
		/**
		 * Creates a strong wrapper reference.
		 *
		 * @param o
		 */
		public WrapperReference(Object o) {
			strong = o;
			weak = null;
		}
		
		/**
		 * Creates a weak wrapper reference.
		 *
		 * @param o
		 */
		public WrapperReference(WeakReference<Object> o) {
			strong = null;
			weak = o;
		}
		
		/**
		 * Gets the wrapper object referenced by this reference.
		 *
		 * @return
		 */
		public Object get() {
			if (null != strong)
				return strong;
			return weak.get();
		}
	}
}
